/*
----------------------------------------------------------------------------------------------------
ToddFun_GE_G35_XmasLights iPhone Controller
By Todd Harrison 12/07/2014 <www.ToddFun.com>
----------------------------------------------------------------------------------------------------
To be used with GE G-35 LED lights, RFduino and the "RFduino ColorWheel" iPhone App
Complete tutorial video and full use instructions at:
http://www.ToddFun.com/????????

I have 25 Christmas patterns programmed which can be selected by setting the r,g,b color sliders to zero 
  then tap the number box for just the "r" slider to bring up your keypad.
  Enter program numbers (1-25) as the "r" color to switch between the 25 programmed patters.
  Or just tap any spot on the color wheel to set a fixed color.
  Entering r,g,b = 0,0,0 starts a demo program cycling one time through all programmed patterns showing each pattern for 10 seconds.
  
Entering r,g,b = 0,0,0 starts a demo program cycling through all programmed patterns showing each pattern for 10 seconds.
  When finished showing each pattern this special demo cycle mode will auto cancel and revert back to the prier saved cycle duration and cycle mode (enabled or disabled).
User can toggle lock mode by entering r,g,b = 55,55,55 or you can program your own lock code so long as the first number is greater than your program count.
  Upon changing the lock mode all current features are saved including lock mode, cycle duration, cycle mode and current color or pattern.
Entering r,g,b = 55,55,56 save all current features including lock mode, cycle duration, cycle mode and current color or pattern.
Entering r,g,b = 55,55,57 initiates program mode
  The next 3 numbers entered for r,g,b sets the standard cycle time in hours, minutes, seconds accumulative.
  So if you enter 55,55,57 followed by 0,2,30 you will have a 2 minute and 30 second cycle time. 
  Standard cycle mode must be enabled to take effect.
  Upon changing the cycle duration the cycle mode is also enabled and all current features are saved including lock mode, cycle duration, cycle mode and current color or pattern.
Standard cycle mode can be started by entering 255,255,255 or by running the demo 0,0,0 and when the demo mode completes standard cycle mode is auto enabled.
If the user enters anything other than 0,0,0 or 255,255,255 and is not entering one of the interactive modes above then cycle mode is cancelled, but this new mode is NOT saved.
A reboot or watchdog reset reverts to saved setting in flash page.

Good tip: Use this "while(RFduinoBLE.radioActive){}" before any statement or process call that could take a bit of CPU cycles.
          It seems the radio needs undivided attention every 80ms for about 6ms
		  If you are doing to much for too long in your code and at the wrong time the radio will hang and only the watchdog will save you with a reset.
		  I have had good luck with this but without it I seem to have many random hangups.

Good tip: I had a lot of commented out print statements that you can un-comment and use for testing while making new programs patterns or other changes.
          There are also commented out print statements at the top of each ToddFun_Pattern_<program name>.cpp file in my "ToddFun_GE_G35_XmasLights_Library" library.
		  You can use these print statements to trace what programs are being called and when from the monitor window of the Arduino IDE.
          It is best to comment out these print calls when finished for smoother performance and less chance of a hang racing with the radio.
		  
Good tip: I have named all my program pattern files prefixed with ToddFun_Pattern_<program name> in my "ToddFun_GE_G35_XmasLights_Library" library for easy identification.
          The files are all renamed with the ToddFun_ prefix so that you can use other people's libraries along side mine without issues.
		  If you start making your own changes to the library files I suggest you do the same just to keep things simple and clearly separated.

My Christmas Program Pattern Picks in ToddFun_Programs.cpp library file (see Readme.md for full list of choices).
1:  Christmas Switching - Snoopy's house in "A Charlie Brown Christmas"
2:  Christmas Twinkle - Purple Orange Green
3:  Christmas Switching - Red Green
4:  Candy Cane Wave
5:  Green Phasing	
6:  Red & Green Chasing
7:  Red to Green Dither	
8:  Red, Green & Blue Chasing
9:  Red, White & Green Chasing
10: Red, Green, Blue & Yellow Chasing
11: Red, Cyan, Green, Magenta, Blue & Yellow Chasing
12: White Twinkle
13: Christmas Starlight
14: Cold White Icicle - Slow Pulse
15: Slow Pulse
16: Full Rainbow	
17: Full Rainbow Interlaced	
18: Autumn Rainbow
19: Colour Wheel Smooth
20: Always Changing
21: Blue Phasing
22: Green to Blue Dither
23: Blue Wave
24: Subtly Changing Pastels
25: Elaine's Christmas Twinkle
		  
My sketch, library and some set-up was based on this original code and content:
    At <www.instructables.com/id/RFduino-GE-G35-Christmas-Lights/>
    and code by Mike Tsao <github.com/sowbug>
    and code by Mark <github.com/MarkEMarkEMark/G35-MEO-Programs>
	and code by Paul Martis (http://www.digitalmisery.com)
*/

#include <RFduinoBLE.h>
#include <ToddFun_G35String.h>
#include <ToddFun_G35StringGroup.h>
#include <ToddFun_ProgramRunner.h>
#include <ToddFun_Programs.h>

// Total # of lights on string (usually 50, 48, or 36). Maximum is 63, because
// the protocol uses 6-bit addressing and bulb #63 is reserved for broadcast messages.
#define LIGHT_COUNT (50)

// RFduino data pin number.
#define G35_PIN (6)

// define watchdog time out trigger
#define WATCHDOG_COUNTER  98304 // = 1 second 32khz ticks * 3

// select a flash page that isn't in use (see Memory.h for more info)
#define  MY_FLASH_PAGE  251

struct saved_data_t
{
  int magic_number;
  bool lock_mode;
  bool cycle_mode;
  uint16_t duration;
  uint8_t red; 
  uint8_t green;
  uint8_t blue;
};

// a flash page is 1K in length, so page 251 starts at address 251 * 1024 = 257024 = 3EC00 hex
saved_data_t *p = (saved_data_t*)ADDRESS_OF_PAGE(MY_FLASH_PAGE);

// default startup save structure with defaults values for use on first uses on a new RFduino.  
//   after the first use on a new RFduino this flash page will retain your prior setting for consecutive reboot or rest
//   you may not want this feature because you are limit to about 10,000 re-flashes of this flash page.
//   every change to default lock mode, cycle duration, color or pattern requires a re-flash.
//   so  you might need a new RFduino ever 2 to 5 years depending on usage.
struct saved_data_t value = {0xCAFEBABE, false, true, 120, 255, 255, 255};

// for testing only
//void dump_data( struct saved_data_t *p )
//{ 
//  Serial.print("  magic_number = ");
//  Serial.println(p->magic_number);
//  Serial.print("  lock_mode = ");
//  Serial.println(p->lock_mode);
//  Serial.print("  cycle_mode = ");
//  Serial.println(p->cycle_mode);
//  Serial.print("  duration = ");
//  Serial.println(p->duration);  
//  Serial.print("  red = ");
//  Serial.println(p->red); 
//  Serial.print("  green = ");
//  Serial.println(p->green);  
//  Serial.print("  blue = ");
//  Serial.println(p->blue);   
//}

const int PROGRAM_COUNT = ToddFun_ProgramGroup::ProgramCount;

// Lights initialisation
ToddFun_G35String lights(G35_PIN, LIGHT_COUNT); 
ToddFun_ProgramGroup programs;
ToddFun_G35StringGroup string_group;

ToddFun_LightProgram* CreateProgram(uint8_t program_index, uint8_t pattern)
{
	return programs.CreateProgram(string_group, program_index, pattern);
}

// How long each program should run in seconds.
// Will only use this if you ensable_time_based_switching()
uint16_t programDuration = 120;

// Used for interactive control modes like setting programDuration
int mode = 0;
bool lockmode = false;
bool startup = true;
uint8_t v_last_red;
uint8_t v_last_green;
uint8_t v_last_blue;
		
// default program you like.
uint8_t program_index_default = 1;
  
ToddFun_ProgramRunner runner(CreateProgram, PROGRAM_COUNT, programDuration);

void flashSave()
{
  while(RFduinoBLE.radioActive){}
  value = {0xCAFEBABE, lockmode, runner.is_switch_time_based(), programDuration, v_last_red, v_last_green, v_last_blue};

  // for testing only
  //while(RFduinoBLE.radioActive){}
  //Serial.print("v_last_red: "); 
  //while(RFduinoBLE.radioActive){}
  //Serial.print(v_last_red);
  //while(RFduinoBLE.radioActive){}
  //Serial.print(" v_last_green; ");
  //while(RFduinoBLE.radioActive){}
  //Serial.print(v_last_green);
  //while(RFduinoBLE.radioActive){}
  //Serial.print(" v_last_blue: ");
  //while(RFduinoBLE.radioActive){}
  //Serial.println(v_last_blue);
		
  // wait until the radio is active which is ever 80ms
  //   this wastes time, but ensures we will get enough time to erase the flash page before the radio back on again.
  while(!RFduinoBLE.radioActive){}

  // this is roughly the amount of time, from when the radio notification triggers to when the radio is inactive
  delay(6);   
  flashPageErase(MY_FLASH_PAGE);
  
  // rinse and repeat but write flash this time.
  while(!RFduinoBLE.radioActive){}
  delay(6);   
  flashWriteBlock(p, &value, sizeof(value)); 
}
 
void setup() {
  //for serial monitor debugging only!
  //Serial.begin(9600);
  
  // Set data output pin for lights
  pinMode(G35_PIN, OUTPUT);
  delay(50);
	
  lights.enumerate();
  string_group.AddString(&lights);
  
  // Start up the RFduino Bluetooth 4.0 Low Energy BLE RF Module and advertise its name.
  RFduinoBLE.deviceName = "ToddFun.com";
  RFduinoBLE.advertisementData = "G35";
  RFduinoBLE.txPowerLevel = +4;
  RFduinoBLE.begin();
  
  delay(50);
  
  // setup the watchdog so if the program locks up the watchdog can reset the RFduino 
  pinMode(2, OUTPUT);
  digitalWrite(2, HIGH);
  delay(2000);
  // init watchdog
  NRF_WDT->CONFIG = (WDT_CONFIG_SLEEP_Run << WDT_CONFIG_SLEEP_Pos);
  NRF_WDT->CRV = WATCHDOG_COUNTER;    
  NRF_WDT->RREN = WDT_RREN_RR0_Enabled << WDT_RREN_RR0_Pos;
  NRF_WDT->TASKS_START = 1;
  digitalWrite(2, LOW);

  // for testing only
  //Serial.println("Darn the watchdog had to do a reset!"); 

  // if flash page is not initialized, initialize it
  if (p->magic_number != 0xCAFEBABE)
    flashSave();	

  //just dump what is in memory so we know what we are storing in flash
  //dump_data(p);

  // Start with default or last saved cycle mode, duration, lock mode and values for last user input of r,g,b.
  if (p->cycle_mode)
  {
    runner.program_duration_seconds_ = p->duration;
	runner.master_program_duration_seconds_ = p->duration;
	runner.enable_time_based_switching();
	runner.next_switch_millis_ = 0;
  }
  else
  {
	runner.disable_time_based_switching();
  }
  
  lockmode = p->lock_mode;
  
  //set to last saved color or pattern
  char myData[3] = {p->red, p->green, p->blue};
  RFduinoBLE_onReceive(myData, 3);
  
  //During startup we don't want to lockout changes.
  startup = false;
}

void loop() {
  while(RFduinoBLE.radioActive){}
  NRF_WDT->RR[0] = WDT_RR_RR_Reload; // Reload watchdog register
  while(RFduinoBLE.radioActive){}
  runner.loop();
}

void RFduinoBLE_onReceive(char *data, int len) {
  // each transmission should contain an RGB triple
  if (len >= 3)
  {
    // get the RGB values
    uint8_t r = data[0];
    uint8_t g = data[1];
    uint8_t b = data[2];
	
	// Interactive mode so the user can program programDuration in the following 3 inputs for seconds, minutes, hours.
	if (mode > 0) 
	{
		if (mode == 1) 
		{
			// we are adding hours
			if (r > 15)
			{
				r = 15; // 15 hours is long enough and our uint16_t can't go much larger in seconds anyway.
			}
			programDuration = int(r)*3600; // convert to seconds
			mode = 2;
			
			FlashMe();
		}
		else if (mode == 2) 
		{
			// we are adding minutes
			if (g > 60)
			{
				g = 60; // only 60 minutes in an hour.
			}
			programDuration += int(g)*60; // convert to seconds		
			mode = 3;
			
			FlashMe();
		}
		else if (mode == 3)
		{
			// we are adding seconds
			if (b > 60)
			{
				b = 60; // only 60 seconds in a minute.
			}
			programDuration += int(b); // convert to seconds					
			
			// set the programDuration in the running program
			while(RFduinoBLE.radioActive){}
			runner.program_duration_seconds_ = programDuration;
			while(RFduinoBLE.radioActive){}
			runner.master_program_duration_seconds_ = programDuration;
			
			// user has finished entering programDuration so disable interactive mode and rapid flash 3 times.
			mode = 0;
			
			FlashMe();
			FlashMe();
			
			// for testing only
			//while(RFduinoBLE.radioActive){}
			//Serial.print("programDuration in seconds: "); 
			//while(RFduinoBLE.radioActive){}
	        //Serial.println(programDuration);
			
			// user set a new switching time so enable time based switching
			while(RFduinoBLE.radioActive){}
			runner.enable_time_based_switching();
			while(RFduinoBLE.radioActive){}
			runner.next_switch_millis_ = 0;
			
			v_last_red = 255;
			v_last_green = 255;
			v_last_blue = 255;
		
			//save this mode to flash
			flashSave();			
		}
		else 
		{
			// Shouldn't get here but if we do then disable interactive mode
			mode = 0;
		}
		
		return;
	} 
	
	// Interactive mode so the user can program features
	if (r == 55 && g != 55)
	{  
		// flash to let user know command was accepted
		FlashMe();
			
		//nothing to do because the user could be starting lock mode
		return;
	}
	else if (r == 55 && g == 55 && b != 55 && b != 56 && b != 57)
	{
		// flash to let user know command was accepted
		FlashMe();
			
		//nothing to do because the user could be starting lock mode
		return;	
	}
	else if (r == 55 && g == 55 && b == 55)
	{
		// user wants to enable or disable lock mode
		lockmode = !lockmode;
		
		if (lockmode)
		{
			// double flash to let user know lights are locked
			FlashMe();
			FlashMe();
		}
		else
		{
			// flash to let user know lights are unlocked
			FlashMe();
		}
		
		//save this mode to flash
		flashSave();
  
		return;		
	}
	else if (r == 55 && g == 55 && b == 56)
	{ 
		// flash to let user know command was accepted
		FlashMe();
		
		//save all current features to flash
		flashSave();
  
		return;		
	}
	else if (r == 55 && g == 55 && b == 57)
	{  
	    // user entered the complete code for interactive mode so now the user can program cycle duration with the following 3 sequential inputs of r,g,b for hours, minutes, seconds.
		mode = 1;
		
		// flash to let user know command was accepted
		FlashMe();
		
		return;
	}		
	
	// check lock mode and ignore user if lock mode is enabled
	if (lockmode && !startup)
	{	
		return;		
	}	

	if (r == 0 && g == 0 && b == 0)
	{
		// user wants to start cycling through programs every 10 seconds until last program
		//  at end of cycle we should switch duration back to default
		while(RFduinoBLE.radioActive){}
		runner.enable_time_based_switching();
		while(RFduinoBLE.radioActive){}
		runner.program_duration_seconds_ = 10;
		while(RFduinoBLE.radioActive){}
		runner.demo_mode_ = true;
		while(RFduinoBLE.radioActive){}
		runner.next_switch_millis_ = 0;
		
		v_last_red = 255;
		v_last_green = 255;
		v_last_blue = 255;
		
		// flash to let user know command was accepted
		FlashMe();
	}
	else if (r == 255 && g == 255 && b == 255)
	{
		// user wants to start standard continues cycle mode
		while(RFduinoBLE.radioActive){}
		runner.enable_time_based_switching();
		while(RFduinoBLE.radioActive){}
		runner.next_switch_millis_ = 0;

		v_last_red = 255;
		v_last_green = 255;
		v_last_blue = 255;
		
		// flash to let user know command was accepted
		FlashMe();	
	}
	else if (r <= PROGRAM_COUNT && g ==0 && b == 0)
 	{
		// user is entering programs number to run so disable time based switching
		while(RFduinoBLE.radioActive){}
		runner.disable_time_based_switching();

		v_last_red = r;
		v_last_green = 0;
		v_last_blue = 0;
				
		// if red <= PROGRAM_COUNT and green = 0 and blue = 0 then switch to the program matching the red number the user sent.
		while(RFduinoBLE.radioActive){}
		runner.switch_program_to(r);
    }
	else
	{
		// Set all lights to fixed color based on the user selection from the color wheel app. 
		// also stop any existing pattern programs
		
		// user is entering codes so disable time based switching
		while(RFduinoBLE.radioActive){}
		runner.disable_time_based_switching();

		v_last_red = r;
		v_last_green = g;
		v_last_blue = b;
				
		// for testing only
		//while(RFduinoBLE.radioActive){}
	    //Serial.print("Lights set to solid color RGB: "); 
		//while(RFduinoBLE.radioActive){}
	    //Serial.print(r);
		//while(RFduinoBLE.radioActive){}
	    //Serial.print(" ");
		//while(RFduinoBLE.radioActive){}
	    //Serial.print(g);
		//while(RFduinoBLE.radioActive){}
	    //Serial.print(" ");
		//while(RFduinoBLE.radioActive){}
	    //Serial.println(b);
	
		// first map each 8bit input value to a 4bit values
		r = map(r, 0, 255, 0, 15);
		g = map(g, 0, 255, 0, 15);
		b = map(b, 0, 255, 0, 15);

		while(RFduinoBLE.radioActive){}
		runner.stop_running_program();
		while(RFduinoBLE.radioActive){}
	    lights.broadcast_intensity(0);
		
		// Send the new color 3 times so all the bulbs get the signal at least once
		// This does not seem to work all the time. I'm still having problems where the old program is still executing when these solid colors are picked.
		// If the user picks a second time it all seem fine from then on.
		while(RFduinoBLE.radioActive){}
		lights.fill_color(0, LIGHT_COUNT, ToddFun_G35::MAX_INTENSITY, COLOR(r, g, b));

	}
  }
}

void FlashMe() {
	while(RFduinoBLE.radioActive){}
	lights.broadcast_intensity(0);
	while(RFduinoBLE.radioActive){}
	delay(10);
	while(RFduinoBLE.radioActive){}
	lights.broadcast_intensity(ToddFun_G35::MAX_INTENSITY);
	while(RFduinoBLE.radioActive){}
	delay(10);
	while(RFduinoBLE.radioActive){}
	lights.broadcast_intensity(0);
	while(RFduinoBLE.radioActive){}
	delay(10);	
	while(RFduinoBLE.radioActive){}
	lights.broadcast_intensity(ToddFun_G35::MAX_INTENSITY);
	while(RFduinoBLE.radioActive){}
	delay(10);	
	while(RFduinoBLE.radioActive){}
	lights.broadcast_intensity(0);
	while(RFduinoBLE.radioActive){}
	delay(10);
	while(RFduinoBLE.radioActive){}
	lights.broadcast_intensity(ToddFun_G35::MAX_INTENSITY);
}